import pymodconn.model_gen_utils as Model_utils
from pymodconn.major_layers import Encoder_class
from pymodconn.major_layers import Decoder_class
from pymodconn.utils_layers import *
import numpy as np
import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras.models import Model
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '1'


class ModelClass():
    def __init__(self, cfg, current_dt):
        self.cfg = cfg
        self.current_dt = current_dt

        if cfg['if_seed']:
            np.random.seed(cfg['seed'])
            tf.random.set_seed(cfg['seed'])
        self.epochs = cfg['epochs']
        self.batch_size = cfg['batch_size']
        self.save_models_dir = cfg['save_models_dir']
        self.save_results_dir = cfg['save_results_dir']
        self.future_data_col = cfg['future_data_col']
        self.n_past = cfg['n_past']
        self.n_features_input = cfg['n_features_input']
        self.all_layers_neurons = cfg['all_layers_neurons']
        self.input_enc_rnn_depth = cfg['input_enc_rnn_depth']
        self.all_layers_dropout = cfg['all_layers_dropout']
        self.n_future = cfg['n_future']
        self.n_features_output = cfg['n_features_output']
        self.optimizer = cfg['optimizer']  # new
        self.SGD_lr = cfg['SGD']['lr']  # new
        self.SGD_mom = cfg['SGD']['momentum']  # new
        self.Adam_lr = cfg['Adam']['lr']
        self.Adam_b1 = cfg['Adam']['b1']
        self.Adam_b2 = cfg['Adam']['b2']
        self.Adam_epsi = cfg['Adam']['epsi']
        self.loss_func = cfg['loss']
        self.model_type_prob = cfg['model_type_prob']
        self.loss_prob = cfg['loss_prob']
        self.q = cfg['quantiles']
        self.control_future_cells = cfg['control_future_cells']

        self.n_features_output_block = 1

        self.n_outputs_lastlayer = 2 if self.loss_prob == 'parametric' else len(
            self.q)

        self.metrics = cfg['metrics']
        self.mha_head = cfg['mha_head']

        self.cfg = cfg
        self.rnn_type = cfg['rnn_type']
        self.dec_attn_mask = cfg['dec_attn_mask']

        self.fit_type = cfg['fit_type']
        self.seq_len = cfg['seq_len']
        self.if_save_model_image = cfg['if_model_image']
        self.if_model_summary = cfg['if_model_summary']

        self.model_size_GB = 0

        # model structure
        self.IFRNN1 = cfg['IFRNN_input']
        self.IFRNN2 = cfg['IFRNN_output']
        self.IFSELF_MHA = cfg['IFSELF_MHA']
        self.IFCASUAL_MHA = cfg['IFCASUAL_MHA']
        self.IFCROSS_MHA = cfg['IFCROSS_MHA']

        self.save_training_history_file = os.path.join(
            self.save_results_dir, '%s.csv' % (self.current_dt))
        self.save_training_history = os.path.join(
            self.save_results_dir, '%s_history.png' % (self.current_dt))
        self.save_hf5_name = os.path.join(
            self.save_models_dir, '%s.h5' % (self.current_dt))
        self.save_modelimage_name = os.path.join(
            self.save_models_dir, '%s_modelimage.png' % (self.current_dt))
        self.save_modelsummary_name = os.path.join(
            self.save_models_dir, '%s_modelsummary.txt' % (self.current_dt))

        if not os.path.exists(cfg['save_models_dir']):
            os.makedirs(cfg['save_models_dir'])

    def build_model(self):
        """
        Full transformer biLSTM 1 single = input > MHA > biLSTM > output
        here Input goes to MHA and then to biLSTM
        """
        timer = Model_utils.Timer()
        timer.start()
        print('[ModelClass] Model Compiling.....')
        self.future_data_col = self.future_data_col - self.control_future_cells + 1

        # input for encoder_past
        encoder_inputs = tf.keras.layers.Input(
            shape=(self.n_past, self.n_features_input), name='encoder_past_inputs')
        
        encoder_outputs_seq, encoder_outputs_allstates = Encoder_class(
            self.cfg, str(1))(encoder_inputs, init_states=None)

        decoder_outputs_list = []
        
        for i in range(1, self.control_future_cells+1):
            locals()[f"decoder_{i}_inputs"] = tf.keras.layers.Input(shape=(self.n_future, self.future_data_col), name=f"decoder_{i}_inputs")
            
            locals()[f"decoder_{i}_outputs"] = Decoder_class(self.cfg, str(i))(locals()[f"decoder_{i}_inputs"],
                                                                                encoder_outputs_seq,
                                                                                encoder_states=encoder_outputs_allstates)
            decoder_outputs_list.append(locals()[f"decoder_{i}_outputs"])
        
        decoder_outputs_all = MERGE_LIST(self.n_features_output)(decoder_outputs_list)

        # print('Probabilistic or non probabilistic changes')
        # If or not using the probabilistic loss, reshape the output to match the shape required by the loss function.
        if self.model_type_prob == 'prob':
            decoder_outputs3 = tf.keras.layers.Dense(
                units=self.n_features_output * self.n_outputs_lastlayer)(decoder_outputs_all)  # this one
            # Reshape the encoder output to match the shape required by the loss function.
            decoder_outputs4 = tf.keras.layers.Reshape(target_shape=(
                self.n_future, self.n_features_output, self.n_outputs_lastlayer))(decoder_outputs3)
            # If using the parametric loss, apply the soft ReLU activation to ensure a positive standard deviation.
            if self.loss_prob == 'parametric':
                decoder_outputs4 = tf.keras.layers.Lambda(function=lambda x: tf.stack(
                    [x[:, :, :, 0], soft_relu(x[:, :, :, 1])], axis=-1))(decoder_outputs4)
        elif self.model_type_prob == 'nonprob':
            decoder_outputs4 = tf.keras.layers.Lambda(
                lambda x: tf.clip_by_value(x, -1, 1))(decoder_outputs_all)
        else:
            raise ValueError(
                'model_type_prob should be either prob or nonprob')

        decoder_inputs = []
        for i in range(1, self.control_future_cells+1):
            decoder_inputs.append(locals()[f"decoder_{i}_inputs"])

        self.model = Model(
            [encoder_inputs, decoder_inputs], decoder_outputs4)

        Model_utils.Build_utils(
            self.cfg, self.current_dt).postbuild_model(self.model)

    def load_model(self, filepath):
        print('[ModelClass_tfp] Loading model from file %s' % filepath)
        # self.model = load_model(filepath)
        self.model.load_weights(filepath)
        # https://stackoverflow.com/a/69663259/6510598

    def forget_model(self):
        del self.model
        K.clear_session()
        print("Everything forgoten....maybe")



import tensorflow as tf

from pymodconn.utils_layers import *

K = tf.keras.backend


class Encoder_class():
	def __init__(self, cfg, enc_or_dec_number):
		self.cfg = cfg
		self.enc_or_dec_number = enc_or_dec_number

		self.IF_RNN = cfg['IFRNN_input']
		self.IF_MHA = cfg['IFSELF_MHA']
		self.MHA_DEPTH = cfg['ENCODER_MHA_DEPTH']
		self.IF_MASK = 0
	
		self.future_data_col = cfg['future_data_col']
		self.n_past = cfg['n_past']
		self.n_features_input = cfg['n_features_input']
		self.all_layers_neurons = cfg['all_layers_neurons']
		self.all_layers_dropout = cfg['all_layers_dropout']
		self.mha_head = cfg['mha_head']
		self.n_future = cfg['n_future']
		self.n_features_output = cfg['n_features_output']
		self.MHA_RNN = cfg['MHA_RNN']

	def __call__(self, input, init_states=None):
		
		encoder_input = tf.keras.layers.Dense(
			self.all_layers_neurons)(input)

		encoder_input = tf.keras.layers.Dropout(
			self.all_layers_dropout/5)(encoder_input)
		output_cell = encoder_input

		if self.MHA_RNN == 0:
			input_cell = output_cell
			rnn_block = RNN_block_class(self.cfg['IFRNN_input'],
										self.cfg,
										'encoder_input',
										self.enc_or_dec_number)
			
			output_cell, output_states = rnn_block(input_cell, init_states=init_states)

		input_cell = output_cell
		for i in range(self.MHA_DEPTH):
			self_mha = MHA_block_class(self.IF_MHA,
								True,
								self.cfg,
								'encoder',
								self.enc_or_dec_number, 
								'self',
								str(i+1))
			
			output_cell = self_mha(input_cell, input_cell)
			input_cell = output_cell

		if self.MHA_RNN == 1:
			input_cell = output_cell
			rnn_block = RNN_block_class(self.cfg['IFRNN_input'],
										self.cfg,
										'encoder_input',
										self.enc_or_dec_number)
			
			output_cell, output_states = rnn_block(input_cell, init_states=init_states)

		output = output_cell
		return output, output_states


class Decoder_class():
	def __init__(self, cfg, enc_or_dec_number):
		self.cfg = cfg
		self.enc_or_dec_number = enc_or_dec_number

		self.MHA_DEPTH = cfg['DECODER_MHA_DEPTH']

		self.future_data_col = cfg['future_data_col']
		self.n_past = cfg['n_past']
		self.n_features_input = cfg['n_features_input']
		self.all_layers_neurons = cfg['all_layers_neurons']
		self.all_layers_dropout = cfg['all_layers_dropout']
		self.mha_head = cfg['mha_head']
		self.n_future = cfg['n_future']
		self.n_features_output = cfg['n_features_output']

		self.merge_states_units = int(
			self.all_layers_neurons/self.cfg['input_enc_rnn_depth'])
		self.merge_states_units = 8 * int(self.merge_states_units/8)		

	def __call__(self, input, input_vk, encoder_states=None):
		
		encoder_input = tf.keras.layers.Dense(
			self.all_layers_neurons)(input)

		encoder_input = tf.keras.layers.Dropout(
			self.all_layers_dropout/5)(encoder_input)
		output_cell = encoder_input

		input_cell = output_cell
		rnn_block_dec_in = RNN_block_class(self.cfg['IFRNN_input'],
									self.cfg,
									"decoder_input",
									self.enc_or_dec_number)
		
		output_cell, decoder_input_states = rnn_block_dec_in(input_cell, init_states=encoder_states)

		input_cell = output_cell

		for i in range(self.MHA_DEPTH):
			casual_mha = MHA_block_class(self.cfg['IFCASUAL_MHA'],
									False,
									self.cfg,
									'decoder',
									self.enc_or_dec_number, 
									'casual',
									str(i+1))        #<------------------here
			output_cell = casual_mha(input_cell, input_cell)
		
			# decoder multi head attention
			skip_connection = output_cell
			input_cell = output_cell
			cross_mha = MHA_block_class(self.cfg['IFCROSS_MHA'],
									True,
									self.cfg,
									'decoder',
									self.enc_or_dec_number, 
									'cross',
									str(i+1))        #<------------------here
			output_cell = cross_mha(input_cell, input_vk)
			input_cell = output_cell


		input_cell = output_cell
		rnn_block_dec_out = RNN_block_class(self.cfg['IFRNN_output'],
											self.cfg,
											"decoder_output",
											self.enc_or_dec_number)

		merged_states =  STATES_MANIPULATION_BLOCK(self.merge_states_units, self.cfg['MERGE_STATES_METHOD'])(encoder_states, decoder_input_states)

		output_cell, output_states = rnn_block_dec_out(input_cell, init_states = merged_states)
		
		if self.cfg['IF_NONE_GLUADDNORM_ADDNORM'] == 1:
			output_cell, _ = GLU_with_ADDNORM(            #---------------> fix this
								output_layer_size=self.all_layers_neurons,
								dropout_rate=self.all_layers_dropout,
								use_time_distributed=False,
								activation=None)(skip_connection, output_cell)
		elif self.cfg['IF_NONE_GLUADDNORM_ADDNORM'] == 2:
			output_cell = ADD_NORM()(skip_connection, output_cell)

		# decoder_future output
		output_cell = tf.keras.layers.TimeDistributed(
			tf.keras.layers.Dense(self.n_features_output))(output_cell)

		return output_cell



class MHA_block_class():
	def __init__(self, IF_MHA, IF_GRN, cfg, enc_or_dec, enc_or_dec_number, self_or_casual_or_crossMHA, mha_depth_index):
		self.cfg = cfg
		self.enc_or_dec = enc_or_dec
		self.enc_or_dec_number = enc_or_dec_number
		self.IF_MHA = IF_MHA
		self.IF_GRN_mha = IF_GRN
		self.mha_depth_index = mha_depth_index
		self.self_or_casual_or_crossMHA = self_or_casual_or_crossMHA
		
		self.mha_head = self.cfg['mha_head']
		self.n_features_input = self.cfg['n_features_input']
		self.all_layers_neurons = self.cfg['all_layers_neurons']	
		self.all_layers_dropout = self.cfg['all_layers_dropout']
		self.IF_GRN = self.cfg['IF_GRN']
		self.dec_attn_mask = self.cfg['dec_attn_mask']


	def __call__(self, input_q, input_kv):
		if self.IF_MHA:
			
			self.mha_layer_name = self.enc_or_dec + '_' + str(self.enc_or_dec_number) + '_'+ str(self.self_or_casual_or_crossMHA) + 'MHA-' + str(self.mha_depth_index)

			if self.self_or_casual_or_crossMHA == 'casual' and self.dec_attn_mask == 1:
				causal_mask = get_causal_attention_mask()(input_q)
			else:
				causal_mask = None
			
			encoder_mha = tf.keras.layers.MultiHeadAttention(
								num_heads = self.mha_head,
								key_dim = self.n_features_input,
								value_dim = self.n_features_input,
								name=self.mha_layer_name)	

			output_cell = encoder_mha(query=input_q,
										key=input_kv,
										value=input_kv,
										attention_mask=causal_mask,
										training=True)

			if self.cfg['IF_NONE_GLUADDNORM_ADDNORM'] == 1:
				output_cell, _ = GLU_with_ADDNORM(            #---------------> fix this
									output_layer_size=self.all_layers_neurons,
									dropout_rate=self.all_layers_dropout,
									use_time_distributed=False,
									activation=None)(input_q, output_cell)
			elif self.cfg['IF_NONE_GLUADDNORM_ADDNORM'] == 2:
				output_cell = ADD_NORM()(input_q, output_cell)
			
			if self.IF_GRN == True and self.IF_GRN_mha == True:
				output_cell, _ = GRN_layer(
								hidden_layer_size = self.all_layers_neurons,
								output_size = self.all_layers_neurons,
								dropout_rate = self.all_layers_dropout,
								use_time_distributed = True,
								activation_layer_type = 'elu')(output_cell)					
		else:
			
			output_cell = input_q
		
		return output_cell



class RNN_block_class():
	def __init__(self, IF_RNN, cfg, location, num):
		self.IF_RNN = IF_RNN
		self.cfg = cfg
		self.location = location
		self.num = num
		self.n_features_input = self.cfg['n_features_input']
		self.all_layers_neurons = self.cfg['all_layers_neurons']	
		self.all_layers_dropout = self.cfg['all_layers_dropout']
		self.IF_GRN = self.cfg['IF_GRN']
	
	def __call__(self, input_cell, init_states=None):
		if self.IF_RNN:
			rnn = rnn_unit(self.cfg,
						rnn_location=self.location,
						num=self.num)
			
			rnn_outputs1 = rnn(input_cell,
								init_states=init_states)
			
			output_cell = rnn_outputs1[0]
			
			rnn_outputs1_allstates = rnn_outputs1[1:]

			if self.cfg['IF_NONE_GLUADDNORM_ADDNORM'] == 0:
				output_cell = linear_layer(self.all_layers_neurons)(output_cell)
			elif self.cfg['IF_NONE_GLUADDNORM_ADDNORM'] == 1:
				output_cell, _ = GLU_with_ADDNORM(            #---------------> fix this
									output_layer_size=self.all_layers_neurons,
									dropout_rate=self.all_layers_dropout,
									use_time_distributed=False,
									activation=None)(input_cell, output_cell)
			elif self.cfg['IF_NONE_GLUADDNORM_ADDNORM'] == 2:
				output_cell = linear_layer(self.all_layers_neurons)(output_cell)
				output_cell = ADD_NORM()(input_cell, output_cell)

			output_states = rnn_outputs1_allstates
			
			if self.cfg['IF_GRN']:
				output_cell, _ = GRN_layer(
									hidden_layer_size=self.all_layers_neurons,
									output_size=self.all_layers_neurons,
									dropout_rate=self.all_layers_dropout,
									use_time_distributed=True,
									activation_layer_type='elu')(output_cell)
		else:
			output_cell = input_cell
			output_states = None
		
		return output_cell, output_states



from keras.utils.layer_utils import count_params
import os
from typing import *

import numpy as np
import tensorflow as tf
from keras.utils.vis_utils import plot_model
import datetime as dt
K = tf.keras.backend


class Timer():
    def __init__(self):
        self.start_dt = None

    def start(self):
        self.start_dt = dt.datetime.now()

    def stop(self):
        end_dt = dt.datetime.now()
        timetaken = end_dt - self.start_dt
        print('Time taken: %s' % (timetaken))
        # print("Current memory (MBs)",get_memory_info('GPU:0')['current'] / 1048576)
        print("")
        return timetaken

class Build_utils():
    """
    This class is used to compile, make sumamry and save as png.
    """

    def __init__(self, cfg, current_dt):
        if cfg['if_seed']:
            np.random.seed(cfg['seed'])
            tf.random.set_seed(cfg['seed'])
        self.epochs = cfg['epochs']
        self.batch_size = cfg['batch_size']
        self.future_data_col = cfg['future_data_col']
        self.n_past = cfg['n_past']
        self.n_features_input = cfg['n_features_input']
        self.all_layers_neurons = cfg['all_layers_neurons']
        self.all_layers_dropout = cfg['all_layers_dropout']
        self.n_future = cfg['n_future']
        self.n_features_output = cfg['n_features_output']
        self.optimizer = cfg['optimizer']  # new
        self.SGD_lr = cfg['SGD']['lr']  # new
        self.SGD_mom = cfg['SGD']['momentum']  # new
        self.Adam_lr = cfg['Adam']['lr']
        self.Adam_b1 = cfg['Adam']['b1']
        self.Adam_b2 = cfg['Adam']['b2']
        self.Adam_epsi = cfg['Adam']['epsi']
        self.loss_func = cfg['loss']
        self.model_type_prob = cfg['model_type_prob']
        self.loss_prob = cfg['loss_prob']
        self.q = cfg['quantiles']
        self.control_future_cells = cfg['control_future_cells']

        self.n_features_output_block = 1

        self.q = np.unique(np.array(self.q))
        if 0.5 not in self.q:
            self.q = np.sort(np.append(0.5, self.q))
        self.n_outputs_lastlayer = 2 if self.loss_prob == 'parametric' else len(
            self.q)

        self.metrics = cfg['metrics']
        self.mha_head = cfg['mha_head']

        self.cfg = cfg
        self.rnn_type = cfg['rnn_type']
        self.dec_attn_mask = cfg['dec_attn_mask']

        self.model_type = cfg['model_type']
        self.fit_type = cfg['fit_type']
        self.seq_len = cfg['seq_len']
        self.if_save_model_image = cfg['if_model_image']
        self.if_model_summary = cfg['if_model_summary']

        # model structure
        self.IFRNN1 = cfg['IFRNN_input']
        self.IFRNN2 = cfg['IFRNN_output']
        self.IFSELF_MHA = cfg['IFSELF_MHA']
        self.IFCASUAL_MHA = cfg['IFCASUAL_MHA']
        self.IFCROSS_MHA = cfg['IFCROSS_MHA']

        self.save_models_dir = cfg['save_models_dir']
        self.save_results_dir = cfg['save_results_dir']
        self.save_training_history_file = os.path.join(
            self.save_results_dir, '%s.csv' % (current_dt))
        self.save_training_history = os.path.join(
            self.save_results_dir, '%s_history.png' % (current_dt))
        self.save_hf5_name = os.path.join(
            self.save_models_dir, '%s.h5' % (current_dt))
        self.save_modelimage_name = os.path.join(
            self.save_models_dir, '%s_modelimage.png' % (current_dt))
        self.save_modelsummary_name = os.path.join(
            self.save_models_dir, '%s_modelsummary.txt' % (current_dt))

    def CVRMSE_Q50_prob_nonparametric(self, y_true, y_pred):
        return K.sqrt(K.mean(K.square(y_pred[:, :, :, 3] - y_true)))/(K.mean(y_true))

    def CVRMSE_Q50_prob_parametric(self, y_true, y_pred):
        return K.sqrt(K.mean(K.square(y_pred[:, :, :, 0] - y_true)))/(K.mean(y_true))

    def CVRMSE_Q50_nonprob(self, y_true, y_pred):
        return K.sqrt(K.mean(K.square(y_pred - y_true)))/(K.mean(y_true))

    def postbuild_model(self, model) -> Tuple[float, int]:

        self.model = model
        if self.optimizer == 'Adam':
            optimizer_model = tf.keras.optimizers.Adam(
                learning_rate=self.Adam_lr,
                beta_1=self.Adam_b1,
                beta_2=self.Adam_b2,
                epsilon=self.Adam_epsi)

        if self.optimizer == 'SGD':
            optimizer_model = tf.keras.optimizers.SGD(
                learning_rate=self.SGD_lr,
                momentum=self.SGD_mom)

        if self.model_type_prob == 'prob':

            if self.loss_prob == "parametric":  # maximum likelihood estimation is applied to estimate mean and stddev
                self.model.compile(loss=parametric_loss,
                                   optimizer=optimizer_model,
                                   metrics=[self.CVRMSE_Q50_prob_parametric],
                                   run_eagerly=True)

            elif self.loss_prob == "nonparametric":  # non parametric approach where quantiles are predicted
                self.model.compile(loss=lambda y_true, y_pred: nonparametric_loss(y_true, y_pred, self.q),
                                   optimizer=optimizer_model,
                                   metrics=[
                                       self.CVRMSE_Q50_prob_nonparametric],
                                   run_eagerly=True)

        elif self.model_type_prob == 'nonprob':
            self.metrics = self.metrics.copy()
            self.metrics.append(self.CVRMSE_Q50_nonprob)
            self.model.compile(loss=self.loss_func,
                               optimizer=optimizer_model,
                               metrics=self.metrics,
                               run_eagerly=True)

        if self.if_model_summary:
            self.model.summary()

        self.trainable_count = count_params(model.trainable_weights)
        '''
        self.GET_MODEL_SIZE_GB = get_model_memory_usage(
            self.batch_size, self.model)
        print('Trainable parameters in the model : %d' % self.trainable_count)
        '''
        with open(self.save_modelsummary_name, 'w') as f:
            self.model.summary(print_fn=lambda x: f.write(x + '\n'), 
                               line_length = 250, 
                               expand_nested=True,
                               show_trainable=True)
        '''
        with open(self.save_modelsummary_name, 'a') as f:
            f.write('_' * 25 + '\n')
            f.write('Model size in GB : %f' % self.GET_MODEL_SIZE_GB)
        '''
        if self.if_save_model_image:
            print('Saving model as %s' % self.save_modelimage_name)
            plot_model(self.model, to_file=self.save_modelimage_name,
                       show_shapes=True, show_layer_names=True, dpi=600, expand_nested=True)
        print("[ModelClass] Building postmodel DONE!! model can be used as self.model")


def get_model_memory_usage(batch_size, model):
    shapes_mem_count = 0
    internal_model_mem_count = 0
    for l in model.layers:
        layer_type = l.__class__.__name__
        if layer_type == 'Model':
            internal_model_mem_count += get_model_memory_usage(batch_size, l)
        single_layer_mem = 1
        out_shape = l.output_shape
        if type(out_shape) is list:
            out_shape = out_shape[0]
        for s in out_shape:
            if s is None:
                continue
            single_layer_mem *= s
        shapes_mem_count += single_layer_mem

    trainable_count = np.sum([K.count_params(p)
                             for p in model.trainable_weights])
    non_trainable_count = np.sum([K.count_params(p)
                                 for p in model.non_trainable_weights])

    number_size = 4.0
    if K.floatx() == 'float16':
        number_size = 2.0
    if K.floatx() == 'float64':
        number_size = 8.0

    total_memory = number_size * \
        (batch_size * shapes_mem_count + trainable_count + non_trainable_count)
    gbytes = np.round(total_memory / (1024.0 ** 3), 3) + \
        internal_model_mem_count
    print('Memory usage for model: {} GB'.format(gbytes))
    return gbytes


def nonparametric_loss(y_true, y_pred, q):
    '''
    Nonparametric loss function, see Section 3.2.1 in the DeepTCN paper.

    # Parameters:
    y_true: tf.Tensor.
            Actual values of target time series, a tensor with shape (n_samples, n_forecast, n_targets) where n_samples is
            the batch size, n_forecast is the decoder length and n_targets is the number of target time series.

    y_pred: tf.Tensor.
            Predicted quantiles of target time series, a tensor with shape (n_samples, n_forecast, n_targets, n_quantiles)
            where n_samples is the batch size, n_forecast is the decoder length, n_targets is the number of target time
            series and n_quantiles is the number of quantiles.

    q: tf.Tensor.
            Quantiles, a 1-dimensional tensor with length equal to the number of quantiles.

    # Returns:
    tf.Tensor.
            Loss value, a scalar tensor.
    '''

    y_true = tf.cast(tf.expand_dims(y_true, axis=3), dtype=tf.float32)
    y_pred = tf.cast(y_pred, dtype=tf.float32)

    q = tf.cast(tf.reshape(q, shape=(1, len(q))), dtype=tf.float32)
    e = tf.subtract(y_true, y_pred)

    L = tf.multiply(q, tf.maximum(0.0, e)) + \
        tf.multiply(1.0 - q, tf.maximum(0.0, - e))

    return tf.reduce_mean(tf.reduce_mean(tf.reduce_sum(L, axis=-1), axis=-1))


def parametric_loss(y_true, params):
    '''
    Parametric loss function, see Section 3.2.2 in the DeepTCN paper.

    # Parameters:
    y_true: tf.Tensor.
            Actual values of target time series, a tensor with shape (n_samples, n_forecast, n_targets) where n_samples is
            the batch size, n_forecast is the decoder length and n_targets is the number of target time series.

    params: tf.Tensor.
            Predicted means and standard deviations of target time series, a tensor with shape (n_samples, n_forecast,
            n_targets, 2) where n_samples is the batch size, n_forecast is the decoder length and n_targets is the
            number of target time series.

    # Returns:
    tf.Tensor.
            Loss value, a scalar tensor.
    '''

    y_true = tf.cast(y_true, dtype=tf.float32)
    params = tf.cast(params, dtype=tf.float32)

    mu = params[:, :, :, 0]
    sigma = params[:, :, :, 1]

    L = 0.5 * tf.math.log(2 * np.pi) + tf.math.log(sigma) + \
        tf.math.divide(tf.math.pow(y_true - mu, 2), 2 * tf.math.pow(sigma, 2))

    return tf.experimental.numpy.nanmean(tf.experimental.numpy.nanmean(tf.experimental.numpy.nansum(L, axis=-1), axis=-1))





import ruamel.yaml
yaml = ruamel.yaml.YAML()

def read_write_yaml(filename, mode, data_yaml):
	# Read the yaml file
	if mode == 'r':
		with open(filename) as file:
			yaml_configs = yaml.load(file)
		return yaml_configs

	# Write to the yaml file
	if mode == 'w':
		with open(filename, 'w') as yamlfile:
			yaml.dump(data_yaml, yamlfile)


def assert_check_configs(configs):
	assert configs['optimizer'] == 'Adam' or configs['optimizer'] == 'SGD', 'Adam must be either Adam or SGD'
	assert configs['model_type_prob'] == 'prob' or configs[
		'model_type_prob'] == 'nonprob', 'model_type_prob must be either prob or nonprob'
	assert configs['loss_prob'] == 'nonparametric' or configs[
		'loss_prob'] == 'parametric', 'loss_prob must be either nonparametric or parametric'
	#assert configs['control_future_cells'] == 1 or configs['control_future_cells'] == 6, 'control_future_cells must be either 1 or 6'
	assert configs['all_layers_neurons'] % 8 == 0 and configs[
		'all_layers_neurons'] >= 8, 'all_layers_neurons must be divisible by 8 and greater than or equal to 8'
	assert configs['mha_head'] % 8 == 0 and configs[
		'mha_head'] >= 8, 'mha_head must be divisible by 8 and greater than or equal to 8'
	assert configs['rnn_type'] in ['LSTM', 'GRU', 'RNN'], 'rnn_type must be either LSTM, GRU or RNN'
	assert configs['input_enc_rnn_depth'] <= 5, 'max depth of RNN units is 5'


def get_configs(config_filename):
	configs = read_write_yaml(config_filename, 'r', None)
	assert_check_configs(configs)
	return configs



from keras.utils.layer_utils import count_params
import os
from typing import *

import numpy as np
import tensorflow as tf
from keras.utils.vis_utils import plot_model
import datetime as dt
K = tf.keras.backend

def soft_relu(x):
	"""
	Soft ReLU activation function, used for ensuring the positivity of the standard deviation of the Normal distribution
	when using the parameteric loss function. See Section 3.2.2 in the DeepTCN paper.
	"""
	return tf.math.log(1.0 + tf.math.exp(x))

class get_causal_attention_mask(tf.keras.layers.Layer):
	def __init__(self):
		super().__init__()

	def call(self, inputs):
		input_shape = tf.shape(inputs)
		batch_size, seq_length, num_feats = input_shape[0], input_shape[1], input_shape[2]
		i = tf.range(seq_length)[:, tf.newaxis]
		j = tf.range(seq_length)
		mask = tf.cast(i >= j, dtype="int32")
		mask = tf.reshape(mask, (1, seq_length, seq_length))
		mult = tf.concat(
			[tf.expand_dims(batch_size, -1),
			 tf.constant([1, 1], dtype=tf.int32)],
			axis=0)
		return tf.tile(mask, mult)

class get_causal_attention_mask1(tf.keras.layers.Layer):
	"""Returns causal mask to apply for self-attention layer.

	Args:
		self_attn_inputs: Inputs to self attention layer to determine mask shape
	"""
	def __init__(self):
		super().__init__()
		

	def call(self, self_attn_inputs):
		len_s = tf.shape(input=self_attn_inputs)[1]
		bs = tf.shape(input=self_attn_inputs)[:1]
		mask = K.cumsum(tf.eye(len_s, batch_shape=bs), 1)
		return mask



class linear_layer(tf.keras.layers.Layer):
	def __init__(self, hidden_layer_size, activation=None, use_time_distributed=False, use_bias=True):
		super().__init__()
		self.use_time_distributed = use_time_distributed
		self.activation = activation
		self.use_bias = use_bias
		self.hidden_layer_size = hidden_layer_size
		self.dense_layer = tf.keras.layers.Dense(self.hidden_layer_size, activation=self.activation, use_bias=self.use_bias)
		self.td_layer = tf.keras.layers.TimeDistributed(self.dense_layer)	

	def call(self, x, training=False):

		if self.use_time_distributed:
			x = self.td_layer(x)
		else:
			x = self.dense_layer(x)
		
		return x



class GRN_layer(tf.keras.layers.Layer):
	"""
	Applies the gated residual network (GRN) as defined in the TFT paper
	
	Adapted from 
	https://github.com/greatwhiz/tft_tf2/blob/HEAD/libs/tft_model.py
	"""
	def __init__(self, hidden_layer_size, output_size, dropout_rate=None, use_time_distributed=False, activation_layer_type='elu'):
		super().__init__()
		self.hidden_layer_size = hidden_layer_size
		self.output_size = output_size
		self.dropout_rate = dropout_rate
		self.use_time_distributed = use_time_distributed
		self.activation_layer_type = activation_layer_type
		self.linear_layer1 = linear_layer(self.hidden_layer_size,
										activation=None,
										use_time_distributed=self.use_time_distributed)
		self.linear_layer2 = linear_layer(self.hidden_layer_size,
										activation=None,
										use_time_distributed=self.use_time_distributed)
		
		self.activation_layer = tf.keras.layers.Activation(self.activation_layer_type)
		
		self.gluwithaddnorm = GLU_with_ADDNORM(self.output_size, self.dropout_rate, self.use_time_distributed, None)


	def call(self, x, training=False):
		skip = x

		hidden = self.linear_layer1(x)
		hidden = self.activation_layer(hidden)
		hidden = self.linear_layer2(hidden)

		grn_output, gate = self.gluwithaddnorm(skip, hidden)
		
		return grn_output, gate


class GLU_with_ADDNORM(tf.keras.layers.Layer):
	def __init__(self, output_layer_size, dropout_rate, use_time_distributed=True, activation=None):
		super().__init__()
		self.output_layer_size = output_layer_size
		self.dropout_rate = dropout_rate
		self.use_time_distributed = use_time_distributed
		self.activation = activation
		self.linear_layer1 = linear_layer(self.output_layer_size,
										activation=None,
										use_time_distributed=self.use_time_distributed)
		self.GLU_layer = GLU_layer(output_layer_size = self.output_layer_size,
						  dropout_rate = self.dropout_rate,
						  use_time_distributed = self.use_time_distributed,
						  activation=None)
		self.ADD_NORM_layer = ADD_NORM()
	
	def call(self, skip, x, training=False):
		x, gate = self.GLU_layer(x)
		
		x = self.linear_layer1(x)
		
		x = self.ADD_NORM_layer(skip, x)
		return x, gate

class ADD_NORM(tf.keras.layers.Layer):
	def __init__(self):
		super().__init__()
		self.add_layer = tf.keras.layers.Add()
		self.norm_layer = tf.keras.layers.LayerNormalization()

	def call(self, skip, x, training=False):
		x = [skip, x]
		tmp = self.add_layer(x)
		tmp = self.norm_layer(tmp)
		return tmp

class GLU_layer(tf.keras.layers.Layer):
	def __init__(self, output_layer_size, dropout_rate, use_time_distributed=True, activation=None):
		super().__init__()
		self.output_layer_size = output_layer_size
		self.dropout_rate = dropout_rate
		self.use_time_distributed = use_time_distributed
		self.activation = activation
		self.dr_layer = tf.keras.layers.Dropout(self.dropout_rate)
		self.dense_layer = tf.keras.layers.Dense(self.output_layer_size, activation=self.activation)
		self.dense_sigmoid_layer = tf.keras.layers.Dense(self.output_layer_size, activation='sigmoid')
		self.multiply_layer = tf.keras.layers.Multiply()
		self.td_layer = tf.keras.layers.TimeDistributed(self.dense_layer)
		self.td_sigmoid_layer = tf.keras.layers.TimeDistributed(self.dense_sigmoid_layer)

	def call(self, x, training=False):
		if self.dropout_rate is not None:
			x = self.dr_layer(x, training=training)

		if self.use_time_distributed:
			activation_layer = self.td_layer(x)
			gated_layer = self.td_sigmoid_layer(x)
		else:
			activation_layer = self.dense_layer(x)
			gated_layer = self.dense_sigmoid_layer(x)

		x, gate = self.multiply_layer([activation_layer, gated_layer]), gated_layer

		return x, gate



class STATES_MANIPULATION_BLOCK():
	""" Merge states of two different RNNs
	Concates the states and then applies a dense layer
	"""

	def __init__(self, d1, states_manipulation_method):
		self.states_manipulation_method = states_manipulation_method
		self.conc = tf.keras.layers.Concatenate()
		self.dense = linear_layer(d1, activation=None, use_time_distributed=False, use_bias=True)
		self.add_layer = tf.keras.layers.Add()
		self.norm_layer = tf.keras.layers.LayerNormalization()

	def __call__(self, x1, x2):
		if x1 == None or x2 == None or x1 == [] or x2 == []:
			return None
		elif self.states_manipulation_method == 1:
			return None
		elif self.states_manipulation_method == 2:
			return x1
		elif self.states_manipulation_method == 3:
			return x2
		elif self.states_manipulation_method == 4:
			all_x = []
			len_x = len(x1)
			for i in range(len_x):
				a = x1[i]
				b = x2[i]
				x = self.conc([a, b])
				x = self.dense(x)
				all_x.append(x)
			return all_x
		elif self.states_manipulation_method == 5:
			all_x = []
			len_x = len(x1)
			for i in range(len_x):
				a = x1[i]
				b = x2[i]
				x = self.add_layer([a, b])
				x = self.dense(x)
				all_x.append(x)
			return all_x
		elif self.states_manipulation_method == 6:
			all_x = []
			len_x = len(x1)
			for i in range(len_x):
				a = x1[i]
				b = x2[i]
				x = self.add_layer([a, b])
				x = self.norm_layer(x)
				x = self.dense(x)
				all_x.append(x)
			return all_x
		elif self.states_manipulation_method == 7:
			all_x = []
			len_x = len(x1)
			for i in range(len_x):
				a = x1[i]
				b = x2[i]
				x = self.add_layer([a, b])
				all_x.append(x)
			return all_x
		elif self.states_manipulation_method == 8:
			all_x = []
			len_x = len(x1)
			for i in range(len_x):
				a = x1[i]
				b = x2[i]
				x = self.add_layer([a, b])
				x = self.norm_layer(x)
				all_x.append(x)
			return all_x		
		else:
			print('only 8 methods defined, returning None')
			return None




class MERGE_LIST(tf.keras.layers.Layer):
	""" Takes a list of tensors and concats them
	Concates the states and then applies a dense layer
	"""
	def __init__(self, d1):
		super().__init__()
		self.conc = tf.keras.layers.Concatenate()
		self.dense1 = tf.keras.layers.Dense(d1+2)
		self.dense2 = tf.keras.layers.Dense(d1)

	def call(self, x1, training=False):
		x = self.conc(x1)
		x = self.dense1(x)
		x = self.dense2(x)
		return x



class rnn_unit():
	def __init__(self, cfg, rnn_location, num):  #rnn_location = 'encoder_input' or 'decoder_input' or 'decoder_output'

		self.all_layers_dropout = cfg['all_layers_dropout']
		self.rnn_type = cfg['rnn_type']
		self.input_enc_rnn_depth = cfg['input_enc_rnn_depth']
		self.input_enc_rnn_bi = cfg['input_enc_rnn_bi']
		self.all_layers_neurons = cfg['all_layers_neurons']
		self.all_layers_neurons_rnn = int(
			self.all_layers_neurons/self.input_enc_rnn_depth)
		self.all_layers_neurons_rnn = 8 * int(self.all_layers_neurons_rnn/8)
		assert self.all_layers_neurons_rnn > 0, 'all_layers_neurons_rnn should be > 0...change all_layers_neurons or input_enc_rnn_depth in config file'
		self.all_layers_dropout = cfg['all_layers_dropout']
		self.rnn_location = rnn_location
		self.cfg = cfg
		self.num = num

	def __call__(self, input_to_layers, init_states=None):
		if self.input_enc_rnn_depth == 1:
			return self.single_rnn_layer(x_input=input_to_layers, init_states=init_states, mid_layer=False, layername_prefix='Only_')
		else:
			x = self.single_rnn_layer(x_input=input_to_layers, init_states=init_states,
									  mid_layer=True, layername_prefix='First_')  # change
						
			if self.cfg['IF_NONE_GLUADDNORM_ADDNORM'] == 1:
				x, _ = GLU_with_ADDNORM(            #---------------> fix this
									output_layer_size=self.all_layers_neurons,
									dropout_rate=self.all_layers_dropout,
									use_time_distributed=False,
									activation=None)(input_to_layers, x)
			elif self.cfg['IF_NONE_GLUADDNORM_ADDNORM'] == 2:
				x = ADD_NORM()(input_to_layers, x)
			
			for i in range(0, self.input_enc_rnn_depth-2):
				x = self.single_rnn_layer(
					x_input=x, init_states=init_states, mid_layer=True, layername_prefix='Mid_%s_' % (i+1))
				

				if self.cfg['IF_NONE_GLUADDNORM_ADDNORM'] == 1:
					x, _ = GLU_with_ADDNORM(            #---------------> fix this
										output_layer_size=self.all_layers_neurons,
										dropout_rate=self.all_layers_dropout,
										use_time_distributed=False,
										activation=None)(input_to_layers, x)
				elif self.cfg['IF_NONE_GLUADDNORM_ADDNORM'] == 2:
					x = ADD_NORM()(input_to_layers, x)


			return self.single_rnn_layer(x_input=x, init_states=init_states, mid_layer=False, layername_prefix='Last_')

	def single_rnn_layer(self, x_input, init_states, mid_layer=False, layername_prefix=None):
		if self.rnn_type == "LSTM":
			RNN_type = tf.keras.layers.LSTM
		elif self.rnn_type == "GRU":
			RNN_type = tf.keras.layers.GRU
		elif self.rnn_type == "RNN":
			RNN_type = tf.keras.layers.SimpleRNN

		if self.rnn_location == "encoder_input":
			self.init_state = None
		elif self.rnn_location == "decoder_input" or self.rnn_location == "decoder_output":
			self.init_state = init_states

		if mid_layer:
			ret_seq = True
			ret_state = False
		else:
			ret_seq = True
			ret_state = True

		if self.input_enc_rnn_bi:
			self.layername = layername_prefix + self.rnn_location + \
				'_' + str(self.num) + '_bi' + self.rnn_type
		else:
			self.layername = layername_prefix + self.rnn_location + \
				'_' + str(self.num) + '_' + self.rnn_type

		if self.input_enc_rnn_bi:
			x_output = tf.keras.layers.Bidirectional(RNN_type(
				self.all_layers_neurons_rnn,
				dropout=self.all_layers_dropout,
				return_sequences=ret_seq,
				return_state=ret_state,
				name=self.layername))(x_input, initial_state=self.init_state)
		else:
			x_output = RNN_type(
				self.all_layers_neurons_rnn,
				dropout=self.all_layers_dropout,
				return_sequences=ret_seq,
				return_state=ret_state,
				name=self.layername)(x_input, initial_state=self.init_state)
		return x_output

